from girder_large_image.girder_tilesource import GirderTileSource

from girder.constants import AssetstoreType
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item

from . import DICOMFileTileSource
from .assetstore import DICOMWEB_META_KEY


class DICOMGirderTileSource(DICOMFileTileSource, GirderTileSource):
    """
    Provides tile access to Girder items with an DICOM file or other files that
    the dicomreader library can read.
    """

    cacheName = 'tilesource'
    name = 'dicom'

    _mayHaveAdjacentFiles = True

    def _getAssetstore(self):
        files = Item().childFiles(self.item, limit=1)
        if not files:
            return None

        assetstore_id = files[0].get('assetstoreId')
        if not assetstore_id:
            return None

        return File()._getAssetstoreModel(files[0]).load(assetstore_id)

    def _getLargeImagePath(self):
        # Look at a single file and see what type of assetstore it came from
        # If it came from a DICOMweb assetstore, then we will use that method.
        assetstore = self._getAssetstore()
        assetstore_type = assetstore['type'] if assetstore else None
        if assetstore_type == getattr(AssetstoreType, 'DICOMWEB', '__undefined__'):
            return self._getDICOMwebLargeImagePath(assetstore)
        else:
            return self._getFilesystemLargeImagePath()

    def _getFilesystemLargeImagePath(self):
        filelist = [
            File().getLocalFilePath(file) for file in Item().childFiles(self.item)
            if self._pathMightBeDicom(file['name'])]
        if len(filelist) > 1:
            return filelist
        filelist = []
        folder = Folder().load(self.item['folderId'], force=True)
        for item in Folder().childItems(folder):
            if len(list(Item().childFiles(item, limit=2))) == 1:
                file = next(Item().childFiles(item, limit=2))
                if self._pathMightBeDicom(file['name']):
                    filelist.append(File().getLocalFilePath(file))
        return filelist

    def _getDICOMwebLargeImagePath(self, assetstore):
        meta = assetstore[DICOMWEB_META_KEY]
        file = Item().childFiles(self.item, limit=1)[0]
        file_meta = file['dicomweb_meta']

        return {
            'url': meta['url'],
            'study_uid': file_meta['study_uid'],
            'series_uid': file_meta['series_uid'],
            # The following are optional
            'qido_prefix': meta.get('qido_prefix'),
            'wado_prefix': meta.get('wado_prefix'),
            'auth': meta.get('auth'),
        }



import json

from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import Resource
from girder.constants import TokenScope
from girder.exceptions import RestException
from girder.models.assetstore import Assetstore
from girder.utility import assetstore_utilities
from girder.utility.model_importer import ModelImporter
from girder.utility.progress import ProgressContext


class DICOMwebAssetstoreResource(Resource):
    def __init__(self):
        super().__init__()
        self.resourceName = 'dicomweb_assetstore'
        self.route('POST', (':id', 'import'), self.importData)

    def _importData(self, assetstore, params):
        """

        :param assetstore: the destination assetstore.
        :param params: a dictionary of parameters including destinationId,
            destinationType, progress, and filters.
        """
        user = self.getCurrentUser()

        destinationType = params.get('destinationType', 'folder')
        if destinationType not in ('folder', 'user', 'collection'):
            msg = f'Invalid destinationType: {destinationType}'
            raise RestException(msg)

        parent = ModelImporter.model(destinationType).load(params['destinationId'], force=True,
                                                           exc=True)

        limit = params.get('limit') or None
        if limit is not None:
            error_msg = 'Invalid limit'
            try:
                limit = int(limit)
            except ValueError:
                raise RestException(error_msg)

            if limit < 1:
                raise RestException(error_msg)

        try:
            search_filters = json.loads(params.get('filters') or '{}')
        except json.JSONDecodeError as e:
            msg = f'Invalid filters: {e}'
            raise RestException(msg)

        progress = self.boolParam('progress', params, default=False)

        adapter = assetstore_utilities.getAssetstoreAdapter(assetstore)

        with ProgressContext(
            progress, user=user, title='Importing DICOM references',
        ) as ctx:
            items = adapter.importData(
                parent,
                destinationType,
                {
                    'limit': limit,
                    'search_filters': search_filters,
                    'auth': None,
                },
                ctx,
                user,
            )

            if not items:
                msg = 'No DICOM objects matching the search filters were found'
                raise RestException(msg)

    @access.admin(scope=TokenScope.DATA_WRITE)
    @autoDescribeRoute(
        Description('Import references to DICOM objects from a DICOMweb server')
        .modelParam('id', 'The ID of the assetstore representing the DICOMweb server.',
                    model=Assetstore)
        .param('destinationId', 'The ID of the parent folder, collection, or user '
               'in the Girder data hierarchy under which to import the files.')
        .param('destinationType', 'The type of the parent object to import into.',
               enum=('folder', 'user', 'collection'),
               required=True)
        .param('limit', 'The maximum number of results to import.',
               required=False, dataType='int')
        .param('filters', 'Any search parameters to filter DICOM objects.',
               required=False, default={})
        .param('progress', 'Whether to record progress on this operation ('
               'default=False)', required=False, default=False, dataType='boolean')
        .errorResponse()
        .errorResponse('You are not an administrator.', 403),
    )
    def importData(self, assetstore, params):
        return self._importData(assetstore, params)



# Cache these so we only look them up once per run
DICOM_TAGS = {}


def dicom_key_to_tag(key):
    if key not in DICOM_TAGS:
        import pydicom
        from pydicom.tag import Tag
        DICOM_TAGS[key] = Tag(pydicom.datadict.tag_for_keyword(key)).json_key

    return DICOM_TAGS[key]



from girder import events
from girder.api.v1.assetstore import Assetstore as AssetstoreResource
from girder.constants import AssetstoreType
from girder.models.assetstore import Assetstore
from girder.utility.assetstore_utilities import setAssetstoreAdapter

from .dicomweb_assetstore_adapter import DICOMWEB_META_KEY, DICOMwebAssetstoreAdapter
from .rest import DICOMwebAssetstoreResource

__all__ = [
    'DICOMWEB_META_KEY',
    'DICOMwebAssetstoreAdapter',
    'load',
]


def createAssetstore(event):
    """
    When an assetstore is created, make sure it has a well-formed DICOMweb
    information record.

    :param event: Girder rest.post.assetstore.before event.
    """
    params = event.info['params']

    if params.get('type') == AssetstoreType.DICOMWEB:
        event.addResponse(Assetstore().save({
            'type': AssetstoreType.DICOMWEB,
            'name': params.get('name'),
            DICOMWEB_META_KEY: {
                'url': params['url'],
                'qido_prefix': params.get('qido_prefix'),
                'wado_prefix': params.get('wado_prefix'),
                'auth_type': params.get('auth_type'),
            },
        }))
        event.preventDefault()


def updateAssetstore(event):
    """
    When an assetstore is updated, make sure the result has a well-formed set
    of DICOMweb information.

    :param event: Girder assetstore.update event.
    """
    params = event.info['params']
    store = event.info['assetstore']

    if store['type'] == AssetstoreType.DICOMWEB:
        store[DICOMWEB_META_KEY] = {
            'url': params['url'],
            'qido_prefix': params.get('qido_prefix'),
            'wado_prefix': params.get('wado_prefix'),
            'auth_type': params.get('auth_type'),
        }


def load(info):
    """
    Load the plugin into Girder.

    :param info: a dictionary of plugin information.  The name key contains the
                 name of the plugin according to Girder.
    """
    AssetstoreType.DICOMWEB = 'dicomweb'
    setAssetstoreAdapter(AssetstoreType.DICOMWEB, DICOMwebAssetstoreAdapter)
    events.bind('assetstore.update', 'dicomweb_assetstore', updateAssetstore)
    events.bind('rest.post.assetstore.before', 'dicomweb_assetstore',
                createAssetstore)

    (AssetstoreResource.createAssetstore.description
        .param('url', 'The base URL for the DICOMweb server (for DICOMweb)',
               required=False)
        .param('qido_prefix', 'The QIDO URL prefix for the server, if needed (for DICOMweb)',
               required=False)
        .param('wado_prefix', 'The WADO URL prefix for the server, if needed (for DICOMweb)',
               required=False)
        .param('auth_type',
               'The authentication type required for the server, if needed (for DICOMweb)',
               required=False))

    info['apiRoot'].dicomweb_assetstore = DICOMwebAssetstoreResource()



import os

import pytest


@pytest.mark.girder()
@pytest.mark.girder_client()
@pytest.mark.plugin('large_image')
@pytest.mark.plugin('dicomweb')
def testDICOMWebClient(boundServer, fsAssetstore, db):
    from pytest_girder.web_client import runWebClientTest

    spec = os.path.join(os.path.dirname(__file__), 'web_client_specs', 'dicomWebSpec.js')
    runWebClientTest(boundServer, spec, 15000)



from girder.plugin import GirderPlugin

from . import assetstore


class DICOMwebPlugin(GirderPlugin):
    DISPLAY_NAME = 'DICOMweb Plugin'
    CLIENT_SOURCE_PATH = 'web_client'

    def load(self, info):
        assetstore.load(info)



from requests.exceptions import HTTPError

from girder.exceptions import ValidationException
from girder.models.file import File
from girder.models.folder import Folder
from girder.models.item import Item
from girder.utility.abstract_assetstore_adapter import AbstractAssetstoreAdapter

from ..dicom_tags import dicom_key_to_tag

DICOMWEB_META_KEY = 'dicomweb_meta'


class DICOMwebAssetstoreAdapter(AbstractAssetstoreAdapter):
    """
    This defines the interface to be used by all assetstore adapters.
    """

    def __init__(self, assetstore):
        super().__init__(assetstore)

    @staticmethod
    def validateInfo(doc):
        # Ensure that the assetstore is marked read-only
        doc['readOnly'] = True

        required_fields = [
            'url',
        ]

        info = doc.get(DICOMWEB_META_KEY, {})

        for field in required_fields:
            if field not in info:
                raise ValidationException('Missing field: ' + field)

        # If these are empty, they need to be converted to None
        convert_empty_fields_to_none = [
            'qido_prefix',
            'wado_prefix',
            'auth_type',
        ]

        for field in convert_empty_fields_to_none:
            if isinstance(info.get(field), str) and not info[field].strip():
                info[field] = None

        # Now, if there is no authentication, verify that we can connect to the server.
        # If there is authentication, we may need to prompt the user for their
        # username and password sometime before here.
        if info['auth_type'] is None:
            study_instance_uid_tag = dicom_key_to_tag('StudyInstanceUID')
            series_instance_uid_tag = dicom_key_to_tag('SeriesInstanceUID')

            client = _create_dicomweb_client(info)
            # Try to search for series. If we get an http error, raise
            # a validation exception.
            try:
                series = client.search_for_series(
                    limit=1,
                    fields=(study_instance_uid_tag, series_instance_uid_tag),
                )
            except HTTPError as e:
                raise ValidationException('Failed to validate DICOMweb server settings: ' + str(e))

            # If we found a series, then test the wado prefix as well
            if series:
                # The previous query should have obtained uids for a specific
                # study and series.
                study_uid = series[0][study_instance_uid_tag]['Value'][0]
                series_uid = series[0][series_instance_uid_tag]['Value'][0]
                try:
                    # Retrieve the metadata of this series as a wado prefix test
                    client.retrieve_series_metadata(
                        study_instance_uid=study_uid,
                        series_instance_uid=series_uid,
                    )
                except HTTPError as e:
                    raise ValidationException('Failed to validate DICOMweb WADO prefix: ' + str(e))

        return doc

    def initUpload(self, upload):
        msg = 'DICOMweb assetstores are import only.'
        raise NotImplementedError(msg)

    def finalizeUpload(self, upload, file):
        msg = 'DICOMweb assetstores are import only.'
        raise NotImplementedError(msg)

    def deleteFile(self, file):
        # We don't actually need to do anything special
        pass

    def downloadFile(self, file, offset=0, headers=True, endByte=None,
                     contentDisposition=None, extraParameters=None, **kwargs):
        # FIXME: do we want to support downloading files? We probably
        # wouldn't download them the regular way, but we could instead
        # use a dicomweb-client like so:
        # instance = client.retrieve_instance(
        #     study_instance_uid=...,
        #     series_instance_uid=...,
        #     sop_instance_uid=...,
        # )
        # pydicom.filewriter.write_file('output_name.dcm', instance)
        msg = 'Download support not yet implemented for DICOMweb files.'
        raise NotImplementedError(
            msg,
        )

    def importData(self, parent, parentType, params, progress, user, **kwargs):
        """
        Import DICOMweb WSI instances from a DICOMweb server.

        :param parent: The parent object to import into.
        :param parentType: The model type of the parent object.
        :type parentType: str
        :param params: Additional parameters required for the import process.
            This dictionary may include the following keys:

            :limit: (optional) limit the number of studies imported.
            :search_filters: (optional) a dictionary of additional search
                filters to use with dicomweb_client's `search_for_series()`
                function.
            :auth: (optional) if the DICOMweb server requires authentication,
                this should be an authentication handler derived from
                requests.auth.AuthBase.

        :type params: dict
        :param progress: Object on which to record progress if possible.
        :type progress: :py:class:`girder.utility.progress.ProgressContext`
        :param user: The Girder user performing the import.
        :type user: dict or None
        :return: a list of items that were created
        """
        if parentType not in ('folder', 'user', 'collection'):
            msg = f'Invalid parent type: {parentType}'
            raise RuntimeError(msg)

        from wsidicom.uid import WSI_SOP_CLASS_UID

        limit = params.get('limit')
        search_filters = params.get('search_filters', {})

        meta = self.assetstore[DICOMWEB_META_KEY]

        client = _create_dicomweb_client(meta, auth=params.get('auth'))

        study_uid_key = dicom_key_to_tag('StudyInstanceUID')
        series_uid_key = dicom_key_to_tag('SeriesInstanceUID')

        # We are only searching for WSI datasets. Ignore all others.
        # FIXME: is this actually working? For the SLIM server at
        # https://imagingdatacommons.github.io/slim/, none of the series
        # report a SOPClassUID, but we still get all results anyways.
        search_filters = {
            'SOPClassUID': WSI_SOP_CLASS_UID,
            **search_filters,
        }
        fields = [
            study_uid_key,
            series_uid_key,
        ]
        if progress:
            progress.update(message='Searching for series...')

        # FIXME: might need to search in chunks for larger web servers
        series_results = client.search_for_series(
            fields=fields, limit=limit, search_filters=search_filters)
        items = []
        for i, result in enumerate(series_results):
            if progress:
                progress.update(total=len(series_results), current=i,
                                message='Importing series...')

            study_uid = result[study_uid_key]['Value'][0]
            series_uid = result[series_uid_key]['Value'][0]

            # Create a folder for the study, and an item for the series
            folder = Folder().createFolder(parent, parentType=parentType,
                                           name=study_uid, creator=user,
                                           reuseExisting=True)
            item = Item().createItem(name=series_uid, creator=user, folder=folder,
                                     reuseExisting=True)

            # Create a placeholder file with the same name
            file = File().createFile(
                name=f'{series_uid}.dcm',
                creator=user,
                item=item,
                reuseExisting=True,
                assetstore=self.assetstore,
                mimeType=None,
                size=0,
                saveFile=False,
            )
            file['dicomweb_meta'] = {
                'study_uid': study_uid,
                'series_uid': series_uid,
            }
            file['imported'] = True
            File().save(file)

            # FIXME: should we return a list of items (like this), or should
            # we return files?
            items.append(item)

        return items


def _create_dicomweb_client(meta, auth=None):
    from dicomweb_client.api import DICOMwebClient
    from dicomweb_client.session_utils import create_session_from_auth

    # Create the authentication session
    session = create_session_from_auth(auth)

    # Make the DICOMwebClient
    return DICOMwebClient(
        url=meta['url'],
        qido_url_prefix=meta.get('qido_prefix'),
        wado_url_prefix=meta.get('wado_prefix'),
        session=session,
    )
